//------------------------------------------------------------------------------
// <copyright file="CryptoProcessor.h" company="LiquidRobotics">
//
// Copyright (c) 2014 Liquid Robotics, Inc.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
// This software is associated with a release.  Liquid Robotics makes no
// representation that the design, interfaces or component distribution will
// remain unchanged in future releases.
//
// </copyright>
//
// <summary>
// 	Declaration of CryptoProcessor class
// </summary>
//
// <owner>Arash Ushani</owner>
//------------------------------------------------------------------------------

#ifndef __CryptoProcessor_H__
#define __CryptoProcessor_H__

#include <nss/pk11pub.h>

//#define CRYPTO_DEBUG	1

namespace LiquidRobotics {
namespace CryptoLib {

class CryptoProcessor
{

public:

	CryptoProcessor();
	~CryptoProcessor();

	void SetKey(unsigned char *newKey, int len);

	bool Encrypt(unsigned char *data, int dataLen, unsigned char *iv, int ivLen, unsigned char *result, int *result_len);
	bool Decrypt(unsigned char *data, int dataLen, unsigned char *iv, int ivLen, unsigned char *result, int *result_len);

	static void GenerateRandomIV(int ivLen, unsigned char *iv);

	static void test();

private:

	static void init_nss();

	static bool nss_initialized;

	PK11SymKey* SymKey;
	PK11SlotInfo* slot;
	CK_MECHANISM_TYPE cipherMech;
};

}
}

#endif
